import collections
import itertools
import re
from typing import DefaultDict, Iterable, Set

import psutil
from loguru import logger

# noinspection PyProtectedMember
from psutil._common import pconn

ProcessFilterPredicate = collections.namedtuple(
    "ProcessFilterPredicate", ["name", "conditions", "recursive"]
)


def _match_process(process: psutil.Process, predicate: ProcessFilterPredicate) -> bool:
    for condition in predicate.conditions:
        name, regex = condition
        value = getattr(process, name)()
        if isinstance(value, int):
            value = str(value)
        elif isinstance(value, (list, tuple)):
            value = " ".join(value)

        if not re.match(regex, value):
            return False

    return True


def filter_net_connections(
    predicates: Iterable[ProcessFilterPredicate],
) -> DefaultDict[str, Set[pconn]]:
    filtered: DefaultDict[str, Set[pconn]] = collections.defaultdict(set)
    for process, predicate in itertools.product(psutil.process_iter(), predicates):
        try:
            if not _match_process(process, predicate):
                continue

            connections = filtered[predicate.name]
            connections.update(process.connections())
        except psutil.NoSuchProcess:
            logger.debug(
                "Process with PID {} died while filtering network connections",
                process.pid,
            )
            continue

        if predicate.recursive:
            for child in process.children(recursive=True):
                try:
                    connections.update(child.connections())
                except psutil.NoSuchProcess:
                    pass

    return filtered



import functools
import shlex
import shutil
import subprocess
from typing import Optional

from loguru import logger

from .exceptions import MissingDependencyError


# Cache the full executable path just in case subprocess.Popen doesn't
@functools.lru_cache(None)
def _which(binary: str) -> Optional[str]:
    return shutil.which(binary)


def run(command: str, **popen_kwargs) -> subprocess.CompletedProcess:
    executable, *arguments = shlex.split(command)
    path = _which(executable)
    if not path:
        raise MissingDependencyError(f"Executable for command: {command!r} not found")

    logger.debug(command)
    return subprocess.run([path] + arguments, **popen_kwargs)



import sys

from loguru import logger

from .cli import get_argument_parser, main as cli_main
from .exceptions import ConfigError, MissingDependencyError


def main() -> None:
    parser = get_argument_parser()
    arguments = parser.parse_args()
    logger.stop(0)
    logger.add(sys.stderr, level=arguments.logging_level)

    # noinspection PyBroadException
    try:
        cli_main(arguments)
    except KeyboardInterrupt:
        logger.info("Aborted")
    except ConfigError as error:
        logger.error("Invalid configuration: {}", error)
    except MissingDependencyError as error:
        logger.error("Missing dependency: {}", error)
    except Exception:
        logger.exception("Unexpected error occurred:")


if __name__ == "__main__":
    main()



import atexit
import collections
import re
import subprocess
from typing import Iterable, Optional, Tuple, Set, Union

import psutil
from loguru import logger

from .utils import run

MIN_RATE = 8
# "TC store rates as a 32-bit unsigned integer in bps internally, so we can specify a
# max rate of 4294967295 bps" (source: `$ man tc`)
MAX_RATE = 4294967295
IFB_REGEX = re.compile(r"ifb\d+")
FILTER_ID_REGEX = re.compile(r"filter .*? fh ([a-z0-9]+::[a-z0-9]+?)(?:\s|$)")
QDISC_ID_REGEX = re.compile(r"qdisc .+? ([a-z0-9]+?):")
CLASS_ID_REGEX = re.compile(
    r"class .+? (?P<qdisc_id>[a-z0-9]+?):(?P<class_id>[a-z0-9]+)"
)

# This ID seems to be fixed for the ingress QDisc
INGRESS_QDISC_PARENT_ID = "ffff:fff1"

QDisc: Tuple[str, int, int] = collections.namedtuple(
    "QDisc", ["device", "id", "root_class_id"]
)


def _clean_up(
    remove_ifb_device: bool = False, shutdown_ifb_device: Optional[str] = None
) -> None:
    logger.info("Cleaning up IFB device")
    if remove_ifb_device:
        run("rmmod ifb")
    if shutdown_ifb_device:
        run(f"ip link set dev {shutdown_ifb_device} down")


def _activate_device(name: str) -> None:
    run(f"ip link set dev {name} up")


def _create_ifb_device() -> str:
    before = set(psutil.net_if_stats())
    run("modprobe ifb numifbs=1")
    after = set(psutil.net_if_stats())

    # It doesn't matter if the created IFB device is ambiguous, any will do
    name = after.difference(before).pop()
    _activate_device(name)
    return name


def _acquire_ifb_device() -> str:
    devices = psutil.net_if_stats()
    for device_name, device in devices.items():
        if not IFB_REGEX.match(device_name):
            continue

        if not device.isup:
            _activate_device(device_name)
            # Deactivate existing IFB device if it wasn't activated
            atexit.register(_clean_up, shutdown_ifb_device=device_name)

        return device_name

    # Clean up IFB device if it was created
    atexit.register(_clean_up, remove_ifb_device=True)
    return _create_ifb_device()


def _find_free_id(ids: Iterable[int]) -> int:
    if not isinstance(ids, set):
        ids = set(ids)

    current = 1
    while current in ids:
        current += 1
    return current


def _get_free_qdisc_id(device: str) -> int:
    process = run(
        f"tc qdisc show dev {device}", stdout=subprocess.PIPE, universal_newlines=True,
    )

    ids = set()
    for line in process.stdout.splitlines():
        match = QDISC_ID_REGEX.match(line)
        if not match:
            logger.warning("Failed to parse line: {!r}", line)
            continue

        id_string = match.group(1)
        try:
            id_ = int(id_string)
        except ValueError:
            # This should only happen for the ingress QDisc
            logger.debug(
                "Failed to parse QDisc ID as base 10 integer on line: {!r}", line
            )
            id_ = int(id_string, 16)

        ids.add(id_)

    return _find_free_id(ids)


def _get_free_class_id(device: str, qdisc_id: int) -> int:
    process = run(
        f"tc class show dev {device}", stdout=subprocess.PIPE, universal_newlines=True,
    )

    ids = set()
    for line in process.stdout.splitlines():
        match = CLASS_ID_REGEX.match(line)
        if not match:
            logger.warning("Failed to parse line: {!r}", line)
            continue

        groups = match.groupdict()
        if int(groups["qdisc_id"]) == qdisc_id:
            ids.add(int(groups["class_id"]))

    return _find_free_id(ids)


def tc_setup(
    device: str,
    download_rate: Union[int, str] = MAX_RATE,
    download_minimum_rate: Union[int, str] = MIN_RATE,
    upload_rate: Union[int, str] = MAX_RATE,
    upload_minimum_rate: Union[int, str] = MIN_RATE,
    default_download_priority: int = 0,
    default_upload_priority: int = 0,
) -> Tuple[QDisc, QDisc]:
    # Set up IFB device
    run(f"tc qdisc add dev {device} handle ffff: ingress")
    ifb_device = _acquire_ifb_device()
    run(
        f"tc filter add dev {device} parent ffff: protocol ip u32 match u32 0 0 action "
        f"mirred egress redirect dev {ifb_device}"
    )

    # Create IFB device QDisc and root class limited at download_rate
    ifb_device_qdisc_id = _get_free_qdisc_id(ifb_device)
    run(f"tc qdisc add dev {ifb_device} root handle {ifb_device_qdisc_id}: htb")
    ifb_device_root_class_id = _get_free_class_id(ifb_device, ifb_device_qdisc_id)
    run(
        f"tc class add dev {ifb_device} parent {ifb_device_qdisc_id}: classid "
        f"{ifb_device_qdisc_id}:{ifb_device_root_class_id} htb rate {download_rate}"
    )

    ingress_qdisc = QDisc(ifb_device, ifb_device_qdisc_id, ifb_device_root_class_id)
    # Create default class that all traffic is routed through that doesn't match any
    # other filter
    ifb_default_class_id = tc_add_htb_class(
        ingress_qdisc, download_rate, download_minimum_rate, default_download_priority,
    )
    run(
        f"tc filter add dev {ifb_device} parent {ifb_device_qdisc_id}: prio 2 protocol "
        f"ip u32 match u32 0 0 flowid {ifb_device_qdisc_id}:{ifb_default_class_id}"
    )

    # Create device QDisc and root class limited at upload_rate
    device_qdisc_id = _get_free_qdisc_id(device)
    run(f"tc qdisc add dev {device} root handle {device_qdisc_id}: htb")
    device_root_class_id = _get_free_class_id(device, device_qdisc_id)
    run(
        f"tc class add dev {device} parent {device_qdisc_id}: classid "
        f"{device_qdisc_id}:{device_root_class_id} htb rate {upload_rate}"
    )
    egress_qdisc = QDisc(device, device_qdisc_id, device_root_class_id)

    # Create default class that all traffic is routed through that doesn't match any
    # other filter
    device_default_class_id = tc_add_htb_class(
        egress_qdisc, upload_rate, upload_minimum_rate, default_upload_priority,
    )
    run(
        f"tc filter add dev {device} parent {device_qdisc_id}: prio 2 protocol ip u32 "
        f"match u32 0 0 flowid {device_qdisc_id}:{device_default_class_id}"
    )

    return ingress_qdisc, egress_qdisc


def tc_add_htb_class(
    qdisc: QDisc,
    ceil: Union[int, str] = MAX_RATE,
    rate: Union[int, str] = MIN_RATE,
    priority: int = 0,
) -> int:
    class_id = _get_free_class_id(qdisc.device, qdisc.id)
    # rate of 1byte/s is the lowest we can specify. All classes added this way should
    # only be allowed to borrow from the parent class, otherwise it's possible to
    # specify a rate higher than the global rate
    run(
        f"tc class add dev {qdisc.device} parent {qdisc.id}:{qdisc.root_class_id} "
        f"classid {qdisc.id}:{class_id} htb rate {rate} ceil {ceil} prio {priority}"
    )
    return class_id


def _get_filter_ids(device: str) -> Set[str]:
    process = run(
        f"tc filter show dev {device}", stdout=subprocess.PIPE, universal_newlines=True,
    )
    ids = set()
    for line in process.stdout.splitlines():
        match = FILTER_ID_REGEX.match(line)
        if match:
            ids.add(match.group(1))

    return ids


def tc_add_u32_filter(qdisc: QDisc, predicate: str, class_id: int,) -> str:
    before = _get_filter_ids(qdisc.device)
    run(
        f"tc filter add dev {qdisc.device} protocol ip parent {qdisc.id}: prio 1 u32 "
        f"{predicate} flowid {qdisc.id}:{class_id}"
    )
    after = _get_filter_ids(qdisc.device)

    difference = after.difference(before)
    if len(difference) > 1:
        logger.warning("Parsed ambiguous filter handle: {}", difference)
    return difference.pop()


def tc_remove_u32_filter(qdisc: QDisc, filter_id: str) -> None:
    run(
        f"tc filter del dev {qdisc.device} parent {qdisc.id}: handle {filter_id} prio "
        "1 protocol ip u32"
    )


def tc_remove_qdisc(device: str, parent: str = "root") -> None:
    run(f"tc qdisc del dev {device} parent {parent}")



class TrafficTollException(Exception):
    pass


class MissingDependencyError(TrafficTollException):
    pass


class DependencyOutputError(TrafficTollException):
    pass


class ConfigError(TrafficTollException):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['traffictoll']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.4.1,<0.5.0', 'psutil>=5.6.7,<6.0.0', 'ruamel.yamL>=0.16.6,<0.17.0']

entry_points = \
{'console_scripts': ['tt = traffictoll.__main__:main']}

setup_kwargs = {
    'name': 'traffictoll',
    'version': '1.4.0',
    'description': 'NetLimiter-like bandwidth limiting and QoS for Linux',
    'long_description': '# TrafficToll\nNetLimiter-like bandwidth limiting and QoS for Linux\n\n# Description\nTrafficToll allows you to limit download and upload bandwidth globally (per interface)\nand per process, even during the process\' runtime. Additionally it also allows you to\nmake use of QoS traffic prioritization for different processes.\n\nThe configuration can be easily adjusted and new limits and priorities applied at any\npoint, as opposed to similar tools which either can only apply fixed global limits to\nthe interface, certain ports, or require you to start the process through them (and thus\nrestart the target process to change the limits).\n\n# Usage\n`# tt device config`\n\nWhere `device` is the interface you want to limit (usually the one you\nconnect to the internet with). For example:\n\n* `# tt enp3s0 night.yaml --delay 0.5` (regular interface, check every\nhalf second for change in networked processes)\n* `# tt tun0 day.yaml --logging-level DEBUG` (VPN interface, adjust\nlogging level to DEBUG)\n\nCurrently TrafficToll works based on a YAML configuration file. The configuration file\nis best explained by example:\n\n```YAML\n# The rate limits for the specified interface. Specifying these values is useful for two\n# things: 1) you want to limit the used bandwidth for the entire interface or 2) you\n# want to make use of traffic prioritization.\n\n# If you want to 1) limit the used bandwidth for the entire interface, simply specify\n# values below your actual speed: the traffic will be shaped in such a way, that it does\n# not exceed the specified numbers.\n\n# If you want to 2) make use of the traffic prioritization feature, these values must be\n# as close as possible to your real speed: if they are too low, traffic prioritization\n# will work, but you are losing part of your bandwidth; if they are too high, traffic\n# prioritization won\'t work as well as it could. I recommend you use some internet speed\n# test you can find online to get an approximation for the correct values.\n\n# If you don\'t want to do 1) or 2), you can omit these values. Bandwidth limiting per\n# application will still work, just traffic prioritization won\'t work as well or\n# entirely.\ndownload: 5mbps\nupload: 1mbps\n\n# Guaranteed download and upload rates for all global traffic that is not shaped as part\n# of a matched process by TrafficToll. The idea here is to leave enough "guaranteed"\n# bandwidth to all applications not defined in "processes", so that they are not starved\n# to a bandwidth, by processes with higher priority, that would cause the other IP to\n# drop the connection. These are the default values, if omitted. Keep in mind that this\n# doesn\'t reserve the bandwidth -- if this traffic is not made use of, it\'s available\n# to processes with higher priority.\ndownload-minimum: 100kbps\nupload-minimum: 10kbps\n\n# A list of processes you want to match and their respective settings\nprocesses:\n  # You can name the process what you want, it is only used to identify it on the CLI\n  # output\n  "Path of Exile":\n    # Adjust the traffic priority to 0 (highest possible, higher integers mean _lower_\n    # priority) to prevent lag and high pings in the game even when we create heavy\n    # traffic otherwise. If these priorities are omitted, they will default to 0: the\n    # same priority for all the traffic on the interface. As soon as you explicitly\n    # specify a priority for a process, the other traffic on the interface will get a\n    # lower priority, so will other processes where you did not explicitly specify\n    # another priority. In this case "Path of Exile" traffic will have a priority of 0,\n    # the highest, and the interface traffic and other processes will have a priority of\n    # 1.\n    download-priority: 0\n    upload-priority: 0\n\n    # Download and upload rate limits can be entirely omitted if you don\'t want to apply\n    # any, in this case traffic will only be prioritized like specified.\n    #download:\n    #upload:\n\n    # The match section. A process is selected when all predicates in the match section\n    # match. Every attribute psutil.Process\n    # (https://psutil.readthedocs.io/en/latest/index.html#psutil.Process) provides on\n    # Linux can be matched on, using regular expressions. Integer attributes will be\n    # treated as strings and list attributes will be joined using a space before\n    # matching. If you want to, you can also specify a regular expression with an OR\n    # operator and match many processes which will all share the specified bandwidth\n    # limit or traffic priority.\n    # If you do not see a line starting with "Shaping traffic for..." with your process\n    # name in it, while it is clearly causing traffic, your match section is failing.\n    # Please make sure it works correctly.\n    match:\n      - name: "PathOfExile_x64"\n\n  Vivaldi:\n    # Additionally specify fixed bandwidth limits for the browser. Setting bandwidth\n    # limits higher than the interface limits will not work. Different processes\n    # borrow available traffic from the interface limits using their specified priority.\n    download: 2500kbps\n    upload: 500kbps\n\n    # Explicitly set a lower download and upload priority to Path of Exile so our\n    # browsing does not cause the game\'s ping to spike. This would have happened\n    # automatically if we omitted it, because we specified a priority for "Path of\n    # Exile" -- we are just doing it here for clarity.\n    download-priority: 1\n    upload-priority: 1\n    match:\n      - exe: /opt/vivaldi/vivaldi-bin\n\n  Discord:\n    # Set Discord\'s traffic priority to the lowest: this means if we create traffic via\n    # either "Path of Exile" or "Vivaldi" it will get priority and Discord\'s latency and\n    # traffic will slow down accordingly\n    download-priority: 2\n    upload-priority: 2\n\n    # Additionally specify fixed bandwidth limits for Discord. Please note that just\n    # because we specified 50% of the interface bandwidth for "Vivaldi" and "Discord"\n    # each, does not mean "Path of Exile" or other processes will be starved for\n    # traffic: Because we omitted download and upload limits for "Path of Exile" 100% of\n    # the interface rate is automatically assumed for that process, in this case\n    # 5mbps/1mbps. Additionally, because "Path of Exile" has a higher priority than\n    # either of the two, in the extreme case that Vivaldi and Discord utilize their\n    # bandwidth limits fully (and thus the entire interface\'s speed), "Path of Exile"\n    # traffic will get priority and traffic for Vivaldi and Discord will be\n    # appropriately reduced.\n    download: 2500kbps\n    upload: 500kbps\n\n    match:\n      - exe: /opt/discord/Discord\n\n  Riot:\n    download-priority: 2\n    upload-priority: 2\n\n    # The process that actually creates network traffic for electron-based applications\n    # is not uniquely identifiable. Instead we match a uniquely identifiable parent\n    # process, in this case "riot-desktop", and set recursive to True. This instructs\n    # TrafficToll to traffic shape the connections of the matched process and all its\n    # descendants\n    recursive: True\n    match:\n      - name: riot-desktop\n\n  JDownloader 2:\n    download: 300kbps\n    # The download-priority and upload-priority if omitted while another process\n    # explicitly specifies them will automatically be the lowest: in this case 2, the\n    # same as "Discord", our lowest priority process.\n\n    # Since the download and upload priority of this process is the lowest, make sure\n    # that its connections don\'t starve when processes with higher priority use up all\n    # the available bandwidth. These are the default values for each process and will be\n    # applied if omitted.\n    download-minimum: 10kbps\n    upload-minimum: 1kbps\n\n    # JDownloader 2 obviously has its own bandwidth limiting, this is just here as an\n    # example to show that matching on something else than the executable\'s name and\n    # path is possible\n    match:\n      - cmdline: .* JDownloader.jar\n```\n\nUnits can be specified in all formats that `tc` supports, namely: bit (with and without\nsuffix), kbit, mbit, gbit, tbit, bps, kbps, mbps, gbps, tbps. To specify in IEC units,\nreplace the SI prefix (k-, m-, g-, t-) with IEC prefix (ki-, mi-, gi- and ti-)\nrespectively.\n\nAll limits and priorities can be omitted, in which case obviously no traffic shaping\nwill be applied. A process is selected when all predicates in the match section match.\nEvery attribute\n[`psutil.Process`](https://psutil.readthedocs.io/en/latest/index.html#psutil.Process)\nprovides on Linux can be matched on, using regular expressions.\n\nWhen you terminate `tt` using Ctrl+C all changes to the traffic scheduling will be\nreverted, allowing you to easily update the config and apply new limits.\n\n# Installation\n`$ pip install traffictoll`\n\n`tt` has to be run as root.\n\n# Screenshots\nBecause a picture is always nice, even for CLI applications:\n\n![](https://i.imgur.com/a3U5Zdt.png)\n',
    'author': 'Chris Braun',
    'author_email': 'cryzed@googlemail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cryzed/TrafficToll',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import argparse
import atexit
import collections
import enum
import time
from typing import Dict, DefaultDict, Set

from loguru import logger
from ruamel.yaml import YAML

from .exceptions import DependencyOutputError, MissingDependencyError
from .net import ProcessFilterPredicate, filter_net_connections
from .speedtest import test_speed
from .tc import (
    MAX_RATE,
    INGRESS_QDISC_PARENT_ID,
    tc_add_htb_class,
    tc_add_u32_filter,
    tc_remove_qdisc,
    tc_remove_u32_filter,
    tc_setup,
)

CONFIG_ENCODING = "UTF-8"
GLOBAL_MINIMUM_DOWNLOAD_RATE = "100kbps"
GLOBAL_MINIMUM_UPLOAD_RATE = "10kbps"
MINIMUM_DOWNLOAD_RATE = "10kbps"
MINIMUM_UPLOAD_RATE = "1kbps"


class _TrafficType(enum.Enum):
    Ingress = enum.auto()
    Egress = enum.auto()


def get_argument_parser() -> argparse.ArgumentParser:
    argument_parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    argument_parser.add_argument(
        "device", help="The network device to be traffic shaped"
    )
    argument_parser.add_argument("config", help="The configuration file")
    argument_parser.add_argument(
        "--delay",
        "-d",
        type=float,
        default=1,
        help="The delay in seconds between checks for changed connections in monitored "
        "processes",
    )
    argument_parser.add_argument(
        "--logging-level",
        "-l",
        choices={"TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"},
        default="INFO",
        help="The logging level",
    )
    argument_parser.add_argument(
        "--speed-test",
        "-s",
        action="store_true",
        help="Automatically determine upload and download speed before start. Make sure"
        ' to run "speedtest --accept-license" beforehand if you are using the official'
        ' "Ookla Speedtest CLI".',
    )
    return argument_parser


def _clean_up(ingress_device: str, egress_device: str) -> None:
    logger.info("Cleaning up QDiscs")
    tc_remove_qdisc(ingress_device)
    tc_remove_qdisc(egress_device)
    tc_remove_qdisc(egress_device, INGRESS_QDISC_PARENT_ID)


# TODO: Check configuration and raise ConfigError
def main(arguments: argparse.Namespace) -> None:
    with open(arguments.config, "r", encoding=CONFIG_ENCODING) as file:
        config = YAML().load(file)

    config_global_download_rate = config.get("download")
    config_global_upload_rate = config.get("upload")
    if arguments.speed_test:
        logger.info("Running speed test...")

        try:
            result = test_speed()
        except MissingDependencyError as error:
            logger.error("Missing dependency: {}", error)
            result = None
        except DependencyOutputError as error:
            logger.error("Dependency output error: {}", error)
            result = None

        if result:
            logger.info(
                "Determined download speed: {}bps, upload speed: {}bps", *result
            )
            config_global_download_rate, config_global_upload_rate = result
        else:
            logger.error(
                "Failed to automatically determine download and upload speed, falling "
                "back to configuration values"
            )

    if config_global_download_rate is None:
        logger.info(
            "No global download rate specified, download traffic prioritization won't "
            "work"
        )
        global_download_rate = MAX_RATE
    else:
        global_download_rate = config_global_download_rate

    if config_global_upload_rate is None:
        logger.info(
            "No global upload rate specified, upload traffic prioritization won't work"
        )
        global_upload_rate = MAX_RATE
    else:
        global_upload_rate = config_global_upload_rate

    # Determine the priority we want the global default classes to have: this is n+1
    # where n is the lowest defined (=highest integer) priority for any processes in the
    # configuration file. Processes that do not explicitly specify a priority will use
    # this default priority and therefore have the same priority as the global default
    # classes
    lowest_priority = -1
    for name, process in (config.get("processes", {}) or {}).items():
        lowest_priority = max(process.get("upload-priority", -1), lowest_priority)
        lowest_priority = max(process.get("download-priority", -1), lowest_priority)
    lowest_priority += 1

    global_download_priority = config.get("download-priority", lowest_priority)
    global_upload_priority = config.get("upload-priority", lowest_priority)

    config_global_download_minimum_rate = config.get("download-minimum")
    global_download_minimum_rate = (
        GLOBAL_MINIMUM_DOWNLOAD_RATE
        if config_global_download_minimum_rate is None
        else config_global_download_minimum_rate
    )
    if config_global_download_rate is not None:
        logger.info(
            "Setting up global class with max download rate: {} (minimum: {}) and "
            "priority: {}",
            global_download_rate,
            global_download_minimum_rate,
            global_download_priority,
        )
    else:
        logger.info(
            "Setting up global class with unlimited download rate (minimum: {}) and "
            "priority: {}",
            global_download_minimum_rate,
            global_download_priority,
        )

    config_global_upload_minimum_rate = config.get("upload-minimum")
    global_upload_minimum_rate = (
        GLOBAL_MINIMUM_UPLOAD_RATE
        if config_global_upload_minimum_rate is None
        else config_global_upload_minimum_rate
    )
    if config_global_upload_rate is not None:
        logger.info(
            "Setting up global class with max upload rate: {} (minimum: {}) and "
            "priority: {}",
            global_upload_rate,
            global_upload_minimum_rate,
            global_upload_priority,
        )
    else:
        logger.info(
            "Setting up global class with unlimited upload rate (minimum: {}) and "
            "priority: {}",
            global_upload_minimum_rate,
            global_upload_priority,
        )

    ingress_qdisc, egress_qdisc = tc_setup(
        arguments.device,
        global_download_rate,
        global_download_minimum_rate,
        global_upload_rate,
        global_upload_minimum_rate,
        global_download_priority,
        global_upload_priority,
    )
    atexit.register(_clean_up, ingress_qdisc.device, egress_qdisc.device)

    process_filter_predicates = []
    class_ids: Dict[_TrafficType, Dict[str, int]] = {
        _TrafficType.Ingress: {},
        _TrafficType.Egress: {},
    }
    for name, process in (config.get("processes", {}) or {}).items():
        # Prepare process filter predicates to match network connections
        conditions = [list(match.items())[0] for match in process.get("match", [])]
        if not conditions:
            logger.warning(
                "No conditions for: {!r} specified, it will never be matched", name
            )
            continue

        predicate = ProcessFilterPredicate(
            name, conditions, process.get("recursive", False)
        )
        process_filter_predicates.append(predicate)

        # Set up classes for download/upload limiting
        config_download_rate = process.get("download")
        config_download_minimum_rate = process.get("download-minimum")
        config_download_priority = process.get("download-priority")
        download_rate = (
            global_download_rate
            if config_download_rate is None
            else config_download_rate
        )
        download_minimum_rate = (
            MINIMUM_DOWNLOAD_RATE
            if config_download_minimum_rate is None
            else config_download_minimum_rate
        )
        download_priority = (
            lowest_priority
            if config_download_priority is None
            else config_download_priority
        )

        config_upload_rate = process.get("upload")
        config_upload_minimum_rate = process.get("upload-minimum")
        config_upload_priority = process.get("upload-priority")
        upload_rate = (
            global_upload_rate if config_upload_rate is None else config_upload_rate
        )
        upload_minimum_rate = (
            MINIMUM_UPLOAD_RATE
            if config_upload_minimum_rate is None
            else config_upload_minimum_rate
        )
        upload_priority = (
            lowest_priority
            if config_upload_priority is None
            else config_upload_priority
        )

        if config_download_rate is not None:
            logger.info(
                "Setting up class for: {!r} with max download rate: {} (minimum: {}) "
                "and priority: {}",
                name,
                download_rate,
                download_minimum_rate,
                download_priority,
            )
            ingress_class_id = tc_add_htb_class(
                ingress_qdisc, download_rate, download_minimum_rate, download_priority,
            )
            class_ids[_TrafficType.Ingress][name] = ingress_class_id
        elif config_download_priority is not None:
            logger.info(
                "Setting up class for: {!r} with unlimited download rate (minimum: {}) "
                "and priority: {}",
                name,
                download_minimum_rate,
                download_priority,
            )
            ingress_class_id = tc_add_htb_class(
                ingress_qdisc, download_rate, download_minimum_rate, download_priority,
            )
            class_ids[_TrafficType.Ingress][name] = ingress_class_id

        if config_upload_rate is not None:
            logger.info(
                "Setting up class for: {!r} with max upload rate: {} (minimum: {}) and "
                "priority: {}",
                name,
                upload_rate,
                upload_minimum_rate,
                upload_priority,
            )
            egress_class_id = tc_add_htb_class(
                egress_qdisc, upload_rate, upload_minimum_rate, upload_priority,
            )
            class_ids[_TrafficType.Egress][name] = egress_class_id
        elif config_upload_priority is not None:
            logger.info(
                "Setting up class for: {!r} with unlimited upload rate (minimum: {}) "
                "and priority: {}",
                name,
                upload_minimum_rate,
                upload_priority,
            )
            egress_class_id = tc_add_htb_class(
                egress_qdisc, upload_rate, upload_minimum_rate, upload_priority,
            )
            class_ids[_TrafficType.Egress][name] = egress_class_id

    port_to_filter_id: Dict[_TrafficType, Dict[int, str]] = {
        _TrafficType.Ingress: {},
        _TrafficType.Egress: {},
    }

    def add_ingress_filter(port: int, class_id: int) -> None:
        filter_id = tc_add_u32_filter(
            ingress_qdisc, f"match ip dport {port} 0xffff", class_id,
        )
        port_to_filter_id[_TrafficType.Ingress][port] = filter_id

    def add_egress_filter(port: int, class_id: int) -> None:
        filter_id = tc_add_u32_filter(
            egress_qdisc, f"match ip sport {port} 0xffff", class_id,
        )
        port_to_filter_id[_TrafficType.Egress][port] = filter_id

    def remove_filters(port: int) -> None:
        ingress_filter_id = port_to_filter_id[_TrafficType.Ingress].get(port)
        if ingress_filter_id:
            tc_remove_u32_filter(ingress_qdisc, ingress_filter_id)
            del port_to_filter_id[_TrafficType.Ingress][port]

        egress_filter_id = port_to_filter_id[_TrafficType.Egress].get(port)
        if egress_filter_id:
            tc_remove_u32_filter(egress_qdisc, egress_filter_id)
            del port_to_filter_id[_TrafficType.Egress][port]

    filtered_ports: DefaultDict[str, Set[int]] = collections.defaultdict(set)
    while True:
        filtered_connections = filter_net_connections(process_filter_predicates)
        for name, connections in filtered_connections.items():
            ports = set(connection.laddr.port for connection in connections)
            ingress_class_id = class_ids[_TrafficType.Ingress].get(name)
            egress_class_id = class_ids[_TrafficType.Egress].get(name)

            # Add new port filters
            new_ports = sorted(ports.difference(filtered_ports[name]))
            if new_ports:
                logger.info(
                    "Shaping traffic for {!r} on local ports {}",
                    name,
                    ", ".join(map(str, new_ports)),
                )
                for port in new_ports:
                    if ingress_class_id:
                        add_ingress_filter(port, ingress_class_id)
                    if egress_class_id:
                        add_egress_filter(port, egress_class_id)

            # Remove old port filters
            freed_ports = sorted(filtered_ports[name].difference(ports))
            if freed_ports:
                logger.info(
                    "Removing filters for {!r} on local ports {}",
                    name,
                    ", ".join(map(str, freed_ports)),
                )
                for port in freed_ports:
                    remove_filters(port)

            filtered_ports[name] = ports

        # Remove freed ports for unmatched processes (process died or predicate
        # conditions stopped matching)
        for name in set(filtered_ports).difference(filtered_connections):
            freed_ports = sorted(filtered_ports[name])
            if freed_ports:
                logger.info(
                    "Removing filters for {!r} on local ports {}",
                    name,
                    ", ".join(map(str, freed_ports)),
                )
                for port in freed_ports:
                    remove_filters(port)
            del filtered_ports[name]

        time.sleep(arguments.delay)



import collections
import enum
import json
import subprocess
from typing import Optional

from .exceptions import DependencyOutputError
from .utils import run

SPEEDTEST_VERSION_COMMAND = "speedtest --version"
OOKLA_SPEEDTEST_COMMAND = "speedtest --format=json"
SIVEL_SPEEDTEST_COMMAND = "speedtest --json"

SpeedTestResult = collections.namedtuple("SpeedTest", ["download_rate", "upload_rate"])


class _SpeedTestProvider(enum.Enum):
    Ookla = enum.auto()
    Sivel = enum.auto()
    Unknown = enum.auto()


# https://www.speedtest.net/apps/cli
def _ookla_speedtest_cli() -> SpeedTestResult:
    process = run(
        OOKLA_SPEEDTEST_COMMAND, stdout=subprocess.PIPE, universal_newlines=True,
    )

    try:
        result = json.loads(process.stdout)
        return SpeedTestResult(
            result["download"]["bandwidth"], result["upload"]["bandwidth"]
        )
    except (json.JSONDecodeError, KeyError):
        raise DependencyOutputError(
            f"Command: {OOKLA_SPEEDTEST_COMMAND!r} returned unrecognized output: "
            f"{process.stdout!r}"
        )


# https://github.com/sivel/speedtest-cli
def _sivel_speedtest_cli() -> SpeedTestResult:
    process = run(
        SIVEL_SPEEDTEST_COMMAND, stdout=subprocess.PIPE, universal_newlines=True
    )

    try:
        result = json.loads(process.stdout)
        return SpeedTestResult(round(result["download"]), round(result["upload"]))
    except (json.JSONDecodeError, KeyError):
        raise DependencyOutputError(
            f"Command: {SIVEL_SPEEDTEST_COMMAND!r} returned unrecognized output: "
            f"{process.stdout!r}"
        )


def _get_speedtest_provider() -> _SpeedTestProvider:
    process = run(
        SPEEDTEST_VERSION_COMMAND, stdout=subprocess.PIPE, universal_newlines=True
    )
    if process.stdout.startswith("Speedtest by Ookla"):
        return _SpeedTestProvider.Ookla
    elif process.stdout.startswith("speedtest-cli"):
        return _SpeedTestProvider.Sivel
    return _SpeedTestProvider.Unknown


def test_speed() -> Optional[SpeedTestResult]:
    speedtest_version = _get_speedtest_provider()

    if speedtest_version is _SpeedTestProvider.Ookla:
        return _ookla_speedtest_cli()
    elif speedtest_version is _SpeedTestProvider.Sivel:
        return _sivel_speedtest_cli()
